#!/usr/bin/env python
# vim: set fileencoding=latin-1 :
#bhmail.py

"""Sends email if a given IP address is being blocked by blockhosts,
and includes all currently blocked IP addresses in the email message.
Will not send email if given IP address is not yet blocked, or if not a
single address is being blocked.

SMTP is required, it is used for sending email; email header data can be
specified as arguments at run-time, or in the same config file as used by
blockhosts. Defaults can be seen by running with the --help argument, if
you have email setup to receive at root@localhost.localdomain, with
SMTP available at localhost, the defaults might be sufficient to make
this work. Otherwise, edit the section named [bhmail] in the blockhosts.cfg
config file.

This script can be run sequentially after each run of blockhosts.py (to
get notification if currently connecting IP address will be blocked in
future), or can be run periodically using cron (to get list of all
blocked addresses).

Example hosts.allow script:
(/usr/bin/blockhosts.py --echo "%c-%s" ; /usr/bin/bhmail.py --check-ip "%h" --echo "%h")

Example cron run:
/usr/bin/bhmail.py --echo "cron"

====
Based on 29/12/06 patch by Erik Ljungstrm      erik [-at-] ibiblio dot 0rg
    http://www.aczoom.com/cms/forums/blockhosts/patch-enabling-email-alerts

====
bhmail.py Script License
This work is hereby released into the Public Domain.
To view a copy of the public domain dedication, visit
http://creativecommons.org/licenses/publicdomain/ or send a letter to
Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.

Author: Avinash Chopde <avinash@acm.org>
Created: February 2007
http://www.aczoom.com/cms/blockhosts/

"""

# script metadata, also used by setup.py
SCRIPT_ID="bhmail"
VERSION="1.0.0"
VERSION_DATE="February 2007"
AUTHOR="Avinash Chopde"
AUTHOR_EMAIL="avinash@acm.org"
URL="http://www.aczoom.com/cms/blockhosts/"
LICENSE="http://creativecommons.org/licenses/publicdomain/"
DESCRIPTION="Send email regarding BlockHosts blocked IP addresses"
LONG_DESCRIPTION=DESCRIPTION

# -------------------------------------------------------------
import syslog
if __name__ == '__main__':
    syslog.openlog(SCRIPT_ID)

#-------------------------------------
import re
import os, os.path, sys, traceback
import copy
import string
import time
import errno
import smtplib
import ConfigParser

try:
    import datetime
    from optparse import OptionParser, OptionGroup
except ImportError, e:
    print >> sys.stderr, "** Missing modules: datetime or optparse\n** Will not work with earlier python versions - 2.3 or later needed.\n\n", e
    raise

try:
    import blockhosts
    from blockhosts import die, debug, warning, info, error, print_always, sort_by_value
except ImportError, e:
    print >> sys.stderr, "** Failed to load main blockhosts.py module or required functions\n", e
    raise

# --------------------------------------
# defaults for parameters follow this order:
# 1 -> use the value provided as an argument in argv[] to this script
# 2 -> if not, then use the value defined in CONFIG_FILE
# 3 -> if not, then use the value hard-coded in this script - HC_OPTIONS

# Defaults, hard-coded options, these values are used last if no args
# and no config file provided
HC_OPTIONS = {
    "HOSTS_BLOCKFILE": blockhosts.HC_OPTIONS["HOSTS_BLOCKFILE"],
    "SMTP_SERVER": "localhost",
    "NOTIFY_ADDRESS": 'root@localhost.localdomain',
    "SMTP_USER": '',
    "SMTP_PASSWD": '',
    "SENDER_ADDRESS": 'BlockHosts <blockhosts@localhost.localdomain>',
}

#######################################################################
class MailMessage:
    """Compose an email message, and then send it
    
    Constructor takes an dict with all mail header info, as well as a
    string specifying subject, and an array of strings specifying body of
    message
    """

    def __init__(self, options, subject, lines):
        # mail header info is in the options object
        self.__address = options.notify_address.replace('\@', '@')
        self.__sender_address = options.sender_address.replace('\@', '@')
        self.__smtp_server = options.smtp_server
        self.__smtp_user = options.smtp_user.replace('\@', '@')
        self.__smtp_passwd = options.smtp_passwd
        # If smtp_user and passwd is empty, no authentication is necessary

        # mail subject (string) and lines (list of strings)
        self.__subject = subject
        self.__lines = lines

    def sendMail(self, dry_run = False):
        if len(self.__address) == 0:
            debug("   ... no email address specified, not sending any mail")
            return
        session = smtplib.SMTP(self.__smtp_server)
        message = "To: " + self.__address
        message += "\nFrom: "+ self.__sender_address
        message += "\nSubject: " + self.__subject + "\n\n"
        message += "\n".join(self.__lines)
        if dry_run:
            print "\n-----", SCRIPT_ID, ": dry-run, email message-------\n"
            print message
            print "-----"
            return

        if len(self.__smtp_user) > 0:
            debug("bhmail: calling SMTP login...")
            session.login(self.__smtp_user, self.__smtp_passwd)
        debug("bhmail: calling SMTP sendmail...")
        smtpresult = session.sendmail(self.__sender_address, self.__address, message)
        if smtpresult:
            errstr = ""
            for recip in smtpresult.keys():
                errstr = """Unable to deliver mail to: %s Server responded: %s %s %s"""\
                         % (recip, smtpresult[recip][0], smtpresult[recip][1], errstr)
                raise smtplib.SMTPException, errstr
        
#######################################################################

def main():
    """Collects args, open block-file, email blocked addresses"""

    args = sys.argv[1:]

    oparser = OptionParser(version=VERSION, description=DESCRIPTION)

    config_options = copy.copy(HC_OPTIONS)
    blockhosts.get_config(config_options, SCRIPT_ID)
    #DEFAULTS
    oparser.set_defaults(verbose=1,
        dry_run=False,
        check_ip="",
        echo="",
        blockfile=config_options["HOSTS_BLOCKFILE"],
        notify_address=config_options["NOTIFY_ADDRESS"],
        sender_address=config_options["SENDER_ADDRESS"],
        smtp_server=config_options["SMTP_SERVER"],
        smtp_user=config_options["SMTP_USER"],
        smtp_passwd=config_options["SMTP_PASSWD"],
        )

    defaults = oparser.get_default_values()

    oparser.add_option("-q", "--quiet",
        action="store_const", const=0, dest="verbose",
        help="Be as quiet as possible - only print out error messages")

    oparser.add_option("-v", "--verbose",
        action="store_const", const=2, dest="verbose",
        help="Be verbose - print errors, warnings, and info messages")

    oparser.add_option("-g", "--debug",
        action="store_const", const=3, dest="verbose",
        help="Be chatty - print out debug level messages also")

    oparser.add_option("--dry-run", action="store_true",
        help="Don't send any email, just print out email body to stdout instead (%s)" % defaults.dry_run)

    oparser.add_option("--check-ip", type="string", metavar="IPADDRESS",
        help="Instead of mailing entire list of blocked address, just send email if given IP address is being blocked (\"%s\")" % defaults.check_ip)

    oparser.add_option("--echo", type="string", metavar="TAG",
        help="Prints TAG on stderr and syslog, may be used to identify a run of %s (%s)" % (SCRIPT_ID, defaults.echo))

    oconfig = OptionGroup(oparser, "Configuration Options",
        """Hard-coded defaults can be overridden by the values in the
config file - if it exists at %s, and those values can be overriden by
using the command-line options.  The current values are shown in () below.
This program uses a section in the block file (%s) to find blockhosts data.
This section is demarcated by two identical marker lines, with this text
- without the quotes: "%s".
Also see the "INSTALL" file in the blockhosts source package for a
detailed example of the block file.
""" % (blockhosts.CONFIG_FILE, defaults.blockfile, blockhosts.HOSTS_MARKER_LINE))

    oconfig.add_option("--blockfile", type="string", metavar="FILE",
        help="Name of hosts-block-file to read (%s)" % defaults.blockfile)

    oconfig.add_option("--notify-address", metavar="ADDRESS",
        help="Address to send notification emails to (\"%s\")" % defaults.notify_address)

    oconfig.add_option("--sender-address", metavar="ADDRESS",
        help="Address appearing as sender address in notification emails (\"%s\")" % defaults.sender_address)

    oconfig.add_option("--smtp-server", metavar="HOST",
        help="SMTP server to send notification emails through (\"%s\")" % defaults.smtp_server)

    oconfig.add_option("--smtp-user", metavar="USERNAME",
        help="If SMTP authentication is required, use this username (\"%s\")" % defaults.smtp_user)
    oconfig.add_option("--smtp-passwd", metavar="PASSWORD",
        help="If SMTP authentication is required, use this password (\"%s\")" % defaults.smtp_passwd)

    oparser.add_option_group(oconfig)

    (options, remaining_args) = oparser.parse_args()

    # set same logging/message level in blockhosts as in this script
    blockhosts.MESSAGE_LEVEL = options.verbose

    info("%s %s started: %s" % (SCRIPT_ID, VERSION, blockhosts.NOW_DATETIME.strftime("%Y-%m-%d %H:%M:%S")))
    if options.echo:
        # force printing echo tag to screen, and to syslog
        syslog.syslog(syslog.LOG_INFO, "echo tag: %s" % options.echo)
        print_always(" ... echo tag: %s" % options.echo)

    debug("Got options:", options)

    if remaining_args:
        warning("ignoring positional arguments - there should be none!", remaining_args)

    debug("Debug mode enabled.")

    #-----------------------------------------------------------------
    # all options/args collected, construct email message

    # trim the check-ip argument, same regex as used to match log lines
    check_host = ''
    if options.check_ip:
        regex = re.compile(r"""(::ffff:)?(?P<host>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})[^\d]*""")
        m = regex.search(options.check_ip)
        if m:
            try:
                check_host = m.group("host")
            except IndexError:
                die("** Input error: check-ip address not an IP address (%s)?\nPattern: (%s)\n" % (options.check_ip, regex))

    # load all currently blocked hosts
    dh = blockhosts.BlockHosts(options.blockfile)
    try:
        dh.load_hosts_blockfile()
    except (blockhosts.MissingMarkerError, blockhosts.SecondMarkerError):
        die("Failed to load blockfile - block-file marker error\n Expected two marker lines in the file,\n somewhere in the middle of the file:\n%s\n%s\n" % (HOSTS_MARKER_LINE, HOSTS_MARKER_LINE))
    except:
        traceback.print_exc()
        die("Failed to load blockfile")

    watched_hosts = dh.get_initial_watched_hosts()
    blocked_hosts = dh.get_initial_deny_hosts()
    found_check_host = False

    # get list of hosts, sorted by datetime, in descending order
    # hosts = sort_by_value(watched_hosts, reverse = True)
    # hosts = watched_hosts.keys() # unsorted, unspecified order

    lines = ["--- list of all blocked hosts--",]
    subject = SCRIPT_ID + ": currently blocked addresses."

    # always include all blocked hosts in email body
    for host in blocked_hosts:
        try:
            data = watched_hosts[host]
        except KeyError:
            error("%s: found blocked IP (%s), but not in watched list (blockhosts.py error?)." % (SCRIPT_ID, host))
            line = " %15s, <no data - possible error in blockhosts.py?>" % host
        else:
            line = " %15s, count %2d, added %s" % (host, data.count, data.datetime.strftime("%Y-%m-%d-%H-%M"))

        lines.append(line)
        if check_host and host == check_host:
            found_check_host = True
            subject = "%s: blocked: %s" % (SCRIPT_ID, line)
            lines.insert(0, subject + "\n")
            info("bhmail: found check-ip -- ", subject)

    lines.append("---")

    if options.check_ip and (not found_check_host):
        info("bhmail: no email, blocked list does not contain check-ip: ", check_host)
    elif len(blocked_hosts) > 0:
        # only send email if check-ip was found in the blocked list or if
        # there was no check-ip argument and there are blocked hosts
        mailer = MailMessage(options, subject, lines)
        mailer.sendMail(options.dry_run)
    else:
        info("bhmail: no email to send.")

    debug("bhmail: all done.")

#######################################################################
if __name__ == '__main__':
    sys.exit(main())
